/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Forecastfox.
 *
 * The Initial Developer of the Original Code is
 * Jon Stritar <jstritar@MIT.EDU>.
 * Portions created by the Initial Developer are Copyright (C) 2005
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * Jon Stritar <jstritar@MIT.EDU>
 * Richard Klein <richwklein@mchsi.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

/* XPCOM Constants */
const CLASS_ID = Components.ID("{27e1bca0-aded-11d9-9669-0800200c9a66}");
const CLASS_NAME = "Forecastfox Profiles Component";
const CONTRACT_ID = "@ensolis.com/forecastfox/profiles;1";
const nsIDOMXPathResult = Components.interfaces.nsIDOMXPathResult;
const nsIFilePicker = Components.interfaces.nsIFilePicker;
const ffIError = Components.interfaces.ffIError;
const ffIProfiles = Components.interfaces.ffIProfiles;

/******************************************************************************
 * ffProfiles Component
 ******************************************************************************/
function ffProfiles()
{
  this.wrappedJSObject = this;
};

ffProfiles.prototype = {
  _batch: false,
  _loading: false,
  _current: null,
  _nextProfile: null,
  _groups: null,
  _rotating: false,
  _xpath: null,
  _xresolve: null,
  _obs: null,
  _doc: null,
  _branch: null,
  _default: null,
  _excludes: null,
  _timer: null,
  _bundle: null,
  _restart: null,
  
  get batch() { return this._batch; },
  get current() { 
    if (!this._current)
      return "";
      
    return this._current;  
  },
  
  start: function()
  {
    this._loading = true;
    
    //set up components
    var pbs = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService);
    var sbs = Components.classes["@mozilla.org/intl/stringbundle;1"].getService(Components.interfaces.nsIStringBundleService);
    this._xpath = Components.classes["@mozilla.org/dom/xpath-evaluator;1"].getService(Components.interfaces.nsIDOMXPathEvaluator);
    this._xresolve = Components.classes["@ensolis.com/forecastfox/resolver;1"].createInstance(Components.interfaces.nsIDOMXPathNSResolver);
    this._obs = Components.classes["@mozilla.org/observer-service;1"].getService(Components.interfaces.nsIObserverService);
    this._timer = Components.classes["@mozilla.org/timer;1"].createInstance(Components.interfaces.nsITimer);
    this._bundle = sbs.createBundle("chrome://forecastfox/locale/forecastfox.properties");    
    this._branch = pbs.getBranch("forecastfox.");
    this._default = pbs.getDefaultBranch("forecastfox.");
    
    this._groups = new Object();
    this._locks = new Object();
    
    //setup excludes
    this._excludes = new Object();
    this._excludes["profile.current"] = true;
    this._excludes["profile.switch.enabled"] = true;
    this._excludes["profile.switch.delay"] = true;
    this._excludes["migrated"] = true;
    this._excludes["pinged"] = true;
    this._excludes["schedule.span"] = true;
    this._excludes["schedule.bump"] = true;
        
    //migrate data
    this._migrate();
   
    //load profile
    if (this._current)
      this.load(this._current);    
    else
      this.load(this._branch.getCharPref("profile.current"));
        
    //lock pref
    this._branch.lockPref("migrated");

    // start rotation
    this.rotate(ffIProfiles.ROTATE_START); 
    
    //add observer to forecastfox:manager
    this._obs.addObserver(this, "forecastfox:manager", false);
    
    this._loading = false;
    return this._restart; 
  },
  
  stop: function()
  {
    try {
      //remove observers
      this._obs = Components.classes["@mozilla.org/observer-service;1"].getService(Components.interfaces.nsIObserverService);
      this._obs.removeObserver(this, "forecastfox:manager");
    } catch(e) {};
    
    //save the current profile
    this.save(this._current);

    //unlock pref
    this._branch.unlockPref("migrated");
    
    //destroy variables
    this._batch = null;
    this._loading = null;
    this._current = null;
    this._groups = null;
    this._nextLoadProfile = null;
    this._rotating = null;        
    this._xpath = null;
    this._xresolve = null;
    this._doc = null;
    this._obs = null;
    this._branch = null;
    this._default = null;
    this._excludes = null; 
    this._bundle = null;
    this._restart = null;
  },
  
  load: function(aName)
  {
    //delay loading of profile if current one is updating
    if (this._lock("load", aName))
      return;
    
    //save the current profile
    if (this._isProfile(this._current))
      this.save(this._current);
        
    this._batch = true;  

    //get profile
    var profiles = this._evalPath(this._doc, "//profiles");
    var profile = this._evalPath(profiles, "./profile[@id='" + aName + "']");

    //load preferences
    var prefs = this._default.getChildList("", {});
    for (var x=0; x<prefs.length; x++) {
      if (this._excludes[prefs[x]])
        continue;
      var pref = this._evalPath(profile, "./pref[@name='" + prefs[x] + "']");
      this._setPref(pref);
    };
    
    //set current
    this._current = aName;   
    
    //notify complete
    this._batch = false;
    if (!this._loading)
      this._obs.notifyObservers(this, "forecastfox:profile", aName);
  },
  
  save: function(aName)
  {
    //return if nothing passed
    if (!aName)
      return;
      
    var profiles = this._evalPath(this._doc, "//profiles");
    var profile = this._evalPath(profiles, "./profile[@id='" + aName + "']");
    
    //check if this is a new profile
    if (profile)
      profile.parentNode.removeChild(profile);

    //create profile
    profile = this._doc.createElement("profile");
    profile.setAttribute("id", aName);
    
    //add preference to it
    var prefs = this._default.getChildList("", {});
    var pref, type;
    for (var x=0; x<prefs.length; x++) {
      if (this._excludes[prefs[x]])
        continue;    
        
      //if we migrated a pref may exists with no value
      //so we wrap this in a try catch
      try {
        type = this._getPrefType(prefs[x]);
        pref = this._doc.createElement("pref");
        pref.setAttribute("name", prefs[x]);
        pref.setAttribute("type", type);
        pref.setAttribute("value", eval("this._branch.get" + type + "Pref('" + prefs[x] + "')"));
        profile.appendChild(pref);
      } catch(e) {};
    };
    
    //append profile to profiles
    profiles.appendChild(profile);

    //flush to disk
    this.flush(this._doc, this._getFile("profiles.xml", "create"));
  },
  
  flush: function(aDoc, aFile)
  {
    var foStream = Components.classes["@mozilla.org/network/file-output-stream;1"].createInstance(Components.interfaces.nsIFileOutputStream);
    var serializer = Components.classes["@mozilla.org/xmlextras/xmlserializer;1"].createInstance(Components.interfaces.nsIDOMSerializer);
    
    //serialize document to string
    var contents = serializer.serializeToString(aDoc);
          
    //write text to file
    var flags = 0x02 | 0x08 | 0x20;
    foStream.init(aFile, flags, 0664, 0);
    foStream.write(contents, contents.length);
    foStream.close();         
  },
  
  rotate: function(aRotate)
  {
    //cancel in timers
    this._timer.cancel();

    //stop rotating    
    if (aRotate == ffIProfiles.ROTATE_STOP)
      this._rotating = false;
    else {
      this._rotating = true;
      this._schedule();
    }
  },

  create: function(aName)
  {
    //delay creating a profile if current one is updating
    if (this._lock("create", aName))
      return;
    
    //save the current profile
    if (this._isProfile(this._current))
      this.save(this._current);
      
    //make sure profile is saved
    this.save(this._current);

    //clone the current profile
    var name = this._stripName(aName);
    var profiles = this._evalPath(this._doc, "//profiles");
    var profile = this._evalPath(profiles, "./profile[@id='" + this._current + "']");
    var to = profile.cloneNode(true);
   
    //set id
    to.setAttribute("id", name);

        
    //copy cache
    this._cacheFiles(profile, to, false);
    this._cachePrefs(profile, to);
    
    //append
    profiles.appendChild(to);
    
    //set current pref which will
    //load this one and flush data to disk
    this._branch.setCharPref("profile.current", name);
        
  },
  
  rename: function(aName, aTo)
  {
    //delay renaming a profile if current one is updating
    if (this._lock("rename", aName, aTo))
      return;
    
    //make sure profile is saved
    this.save(aName);

    //clone the from profile
    var name = this._stripName(aTo);
    var profiles = this._evalPath(this._doc, "//profiles");
    var profile = this._evalPath(profiles, "./profile[@id='" + aName + "']");
    var to = profile.cloneNode(true);
   
    //set id
    to.setAttribute("id", name);
        
    //copy cache
    this._cacheFiles(profile, to, true);
    this._cachePrefs(profile, to);
    
    //append
    profiles.appendChild(to); 
    
    //remove from
    profiles.removeChild(profile);
    
    //set current pref which will
    //load this one and flush data to disk
    this._current = null;    
    this._branch.setCharPref("profile.current", name);  
  },
  
  remove: function(aName, aTo)
  {
    if (this._lock("remove", aName, aTo))
      return;
      
    //make sure profile is saved
    this.save(aName);
      
    //get the from profile
    var profiles = this._evalPath(this._doc, "//profiles");
    var profile = this._evalPath(profiles, "./profile[@id='" + aName + "']");
    
    //delete cache files
    this._cacheFiles(profile, null, true);
    
    //remove from
    profiles.removeChild(profile);
    
    //set current pref which will
    //load this one and flush data to disk
    this._current = null;
    this._branch.setCharPref("profile.current", aTo);    
  },
  
  getProfiles: function(aCount)
  {
    var profiles = this._evalPath(this._doc, "//profiles");
    var val = new Array();

    profiles = profiles.getElementsByTagName("profile");  
    for (var i=0; i<profiles.length; i++)
      val.push(profiles[i].getAttribute("id"));
    
    val.sort();  
    aCount.value = val.length;
     
    return val;    
  },
  
  open: function(aOpen)
  {
    //set variables
    var alerts = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService(Components.interfaces.nsIPromptService);
    var windowManager = Components.classes["@mozilla.org/appshell/window-mediator;1"].getService().QueryInterface(Components.interfaces.nsIWindowMediator);
    var topWindow = windowManager.getMostRecentWindow(null);
    var doc, profiles, node, current, prefs, pref, type, parent, version, x;
        
    //get file
    var file = this._filePicker(aOpen, topWindow);
    if (!file)
      return;
      
    switch (aOpen) {
      case ffIProfiles.OPEN_EXPORT:
        //copy profiles to export document
        doc = this._getDoc(this._getExport());
        profiles = this._evalPath(this._doc, "//profiles");
        node = doc.getElementsByTagName("forecastfox-settings")[0];
        node.appendChild(profiles.cloneNode(true));
    
        //create a current node
        current = doc.createElement("current");
        
        //add preference to it
        prefs = this._default.getChildList("", {});
        for (x=0; x<prefs.length; x++) {
          if (!this._excludes[prefs[x]])
            continue;    

          try {
            type = this._getPrefType(prefs[x]);
            pref = this._doc.createElement("pref");
            pref.setAttribute("name", prefs[x]);
            pref.setAttribute("type", type);
            pref.setAttribute("value", eval("this._branch.get" + type + "Pref('" + prefs[x] + "')"));
            current.appendChild(pref);
          } catch(e) {};
        };
    
        //append current
        node.appendChild(current);
        
        //flush to disk
    	this.flush(doc, file);
    
        //finished
        alerts.alert(topWindow, this._bundle.GetStringFromName("ff.alert"), this._bundle.GetStringFromName("ff.profiles.export.success"));
        break;
      case ffIProfiles.OPEN_IMPORT:
        //get export document        
        doc = this._getDoc(file);
        if (!this._validDoc(doc)) {
          alerts.alert(topWindow, this._bundle.GetStringFromName("ff.alert"), this._bundle.GetStringFromName("ff.profiles.import.invalid"));
          return;
        };
        
        //have current and profiles
        current = doc.getElementsByTagName("current")[0];
        profiles = doc.getElementsByTagName("profiles")[0];
        if (!current || !profiles) {
          alerts.alert(topWindow, this._bundle.GetStringFromName("ff.alert"), this._bundle.GetStringFromName("ff.profiles.import.invalid"));
          return;
        };
        
        //get version we are importing from
        version = this._evalPath(doc, "//@version");
        if (!version) {
          alerts.alert(topWindow, this._bundle.GetStringFromName("ff.alert"), this._bundle.GetStringFromName("ff.profiles.import.invalid"));
          return;        
        } else
          version = version.textContent;
          
        //get profiles node in current doc
        //and replace with profiles node in import doc
        node = this._evalPath(this._doc, "//profiles");
        parent = node.parentNode;
        parent.removeChild(node);
        parent.appendChild(profiles.cloneNode(true));
        
        //flush data to disk
        this.flush(this._doc, this._getFile("profiles.xml", "create"));
        
        //tranform document if version is current
        if (version != "0.7.7") {
          this._batch = true;
          this._restart = this._createError(ffIError.SEVERITY_INFO, null, "", "");
          this._doc = this._transformDoc(version, false);
          this._batch = false;
          if (this._restart && this._restart.severity != ffIError.SEVERITY_INFO) {
            alerts.alert(topWindow, this._bundle.GetStringFromName("ff.alert"), this._bundle.GetStringFromName("ff.profiles.import.invalid"));
            return;                  
          };
        };
        
        //now apply values from current node
        this._batch = true;
        prefs = current.getElementsByTagName("pref");
        for (x=0; x<prefs.length; x++) {
          if (!this._excludes[prefs[x].getAttribute("name")])
            continue;    
          this._setPref(prefs[x]);
        };
        this._batch = false;
        
        //flush data to disk 
        this.flush(this._doc, this._getFile("profiles.xml", "create"));
        
        //load profile
        this.load(this._branch.getCharPref("profile.current"));
        alerts.alert(topWindow, this._bundle.GetStringFromName("ff.alert"), this._bundle.GetStringFromName("ff.profiles.import.success"));        
        break;
    }
  },
    
  _createError: function(aSeverity, aName, aDescription, aTooltip)
  {
    var description = (aName) ? this._bundle.GetStringFromName(aName + ".label") : aDescription;
    var tooltip = (aName) ? this._bundle.GetStringFromName(aName + ".tooltip") : aTooltip;
    var error = Components.classes["@ensolis.com/forecastfox/error;1"].createInstance(Components.interfaces.ffIError);
    error.init(aSeverity, "@ensolis.com/forecastfox/profiles;1", description, tooltip); 
    return error;
  },
    
  _lock: function(aType, aName, aTo)
  {
    //delay renaming a profile if current one is updating
    for (var i in this._groups) {
      if (this._groups[i].loading) {
        //can't procede - still loading - try to get lock
        for (var j in this._locks) {
          if (this._locks[j]) {
            return true;
          }
        }
        //no locks - take the lock
        this._locks[aType] = new Object();
        this._locks[aType].name = aName;
        if (aTo)
          this._locks[aType].to = aTo;
        return true;
      }
    }
    return false;
  },
  
  _releaseLocks: function()
  {
    var type = null;
    var name = null;
    var to = null;
    
    // find lock
    for (var i in this._locks) {
      if (this._locks[i]) {
        type = i;
        name = this._locks[i].name;
        to = this._locks[i].to;
        break;
      }
    }
    
    // reset locks
    this._locks = new Object();
    
    if (type == null) return;
    
    switch (type) {
      case "load":
        this.load(name);
        break;
      case "rename":
        this.rename(name,to);
        break;
      case "create":
        this.create(name);
        break;
      case "remove":
        this.remove(name,to);
        break;
    }    
  },
  
  
  _filePicker: function(aOpen, aWindow)
  {
    var picker = Components.classes["@mozilla.org/filepicker;1"].createInstance(nsIFilePicker);  
    switch (aOpen) {
      case ffIProfiles.OPEN_EXPORT:
        picker.init(aWindow, this._bundle.GetStringFromName("ff.profiles.export"), nsIFilePicker.modeSave);
        picker.appendFilters(nsIFilePicker.filterXML);
        picker.defaultExtension = ".xml";
        break;
      case ffIProfiles.OPEN_IMPORT:
        picker.init(aWindow, this._bundle.GetStringFromName("ff.profiles.import"), nsIFilePicker.modeOpen);
        picker.appendFilters(nsIFilePicker.filterXML);
        picker.defaultExtension = ".xml";
        break;
    };
    
    // get the file and its contents
    var res = picker.show();
    if (res == nsIFilePicker.returnCanel)
      return null;
    else
      return picker.file;
  },
  
  _isProfile: function(aName)
  {
    var profile = this._evalPath(this._doc, "//profiles/profile[@id='"+aName+"']");
    return (profile != null);
  },
  
  _schedule: function()
  {
    if (!this._rotating)
      this.rotate(ffIProfiles.ROTATE_STOP);
    else {  
      //see if we have multiple profiles
      var profiles = this._evalPath(this._doc, "//profiles");
      if (profiles.getElementsByTagName("profile").length <= 1)
        return;

      var rotate = this._branch.getBoolPref("profile.switch.enabled");
      var rotateTime = this._branch.getIntPref("profile.switch.delay");
    
      //bail out if its disabled
      if (rotate == false || rotateTime == 0)
        return;

      rotateTime *= 1000 * 60;

      // then schedule the timer
      this._timer.initWithCallback(this, rotateTime, Components.interfaces.nsITimer.TYPE_ONE_SHOT);
    }     
  },
    
  _setPref: function(aNode)
  {
    if (!aNode)
      return;
      
    //get variables
    var pref  = aNode.getAttribute("name");
    var type  = aNode.getAttribute("type");
    var value = aNode.getAttribute("value");
    
    //set value based on type
    if (type == "Bool")
      value = (value == "true") ? true : false;
    else if (type == "Int")
      value = Number(value);

    //get the old value
    var oldValue = null;
    var defaultValue = null;
    var hasValue = false;
    try {
      oldValue = eval("this._branch.get" + type + "Pref('" + pref + "')");
      defaultValue = eval("this._default.get" + type + "Pref('" + pref + "')");
      hasValue = eval("this._branch.prefHasUserValue('" + pref + "')");
    } catch(e) {};
    
    //do nothing if the same value
    if (value == oldValue)
      return;
    
    //clear if same as default value
    if (value == defaultValue && hasValue)
      eval("this._branch.clearUserPref('" + pref + "')");
    else {
      if (type == "Char")
        value = "'" + value + "'"; 
      eval("this._branch.set" + type + "Pref('" + pref + "', " + value + ")");
    }
  },
    
  _getPrefType: function(aName)
  {
    //convert integer type to character type
    var type = this._branch.getPrefType(aName);
    var val = null;
    switch (type) {
      case 0:
        val = "";
        break;
      case 32:
      default:
        val = "Char";
        break;
      case 64:
        val = "Int";
        break;
      case 128:
        val = "Bool";
        break;
    };
    
    return val;
  },
    
  _evalPath: function(aNode, aExpr)
  {
    var result = null;
     
    try {   
      var results = this._xpath.evaluate(aExpr, aNode, this._xresolve, nsIDOMXPathResult.ANY_TYPE, null);
      result = results.iterateNext(); 
    } catch(e) {};
    
    return result;    
  },
 
  _migrate: function()
  {
    //set empty error
    this._restart = this._createError(ffIError.SEVERITY_INFO, null, "", "");
             
    var firstTime = false;
    var from = this._branch.getCharPref("migrated");
    var file = null;
    
    //determine if this is the first load
    if (!this._branch.prefHasUserValue("migrated"))
      firstTime = true;
  
    //if this is not the first time we are 
    //installing over a version 0.7 or above
    if (!firstTime) {
      switch (from) {
        case "*":
          // completes the uninstallation/upgrade of weatherfox
          this._removeBranch("weatherfox.", false);
          this._removeDir("weatherfox", false);
          this._doc = this._loadDoc();                   
          break;
        case "0.7.7":
          //same version do nothing
          this._doc = this._loadDoc();
          break;
        default:
          //version 0.7 or above, but below current
          this._doc = this._transformDoc(from, false);
          break;        
      };
    } else {
    
      //we are at a version below 0.7
      var isToolkit = true;
      var isWeatherfox = false;
      var weatherfox = new Array();
      
      //get the extension manager
      try {
        var em = Components.classes["@mozilla.org/extensions/manager;1"].getService(Components.interfaces.nsIExtensionManager);
        var nsIUpdateItem = Components.interfaces.nsIUpdateItem;
      } catch (e) {
         isToolkit = false;
      };
      
      //get profile file
      file = this._getFile("profiles.xml", "exists");
                
      //non toolkit version were not available before 0.7
      if (!isToolkit) {
        if (!file)
          this._doc = this._loadDoc();
        else {
         // bug in version 0.7.0 and 0.7.1 for non toolkit did not update migrate pref
         // so manually override it here if its null
         if (from == "")
           from = "0.7.1";
         this._doc = this._transformDoc(from, false);
        }
      } else {
     
        //determine if we are coming from weatherfox
        try {
          weatherfox[0] = em.getItemForID("{BA8E053E-2867-4772-B9F0-26A5979AFA09}");
        } catch(e) {
          weatherfox = em.getItemList("{BA8E053E-2867-4772-B9F0-26A5979AFA09}", nsIUpdateItem.TYPE_EXTENSION, {});
        };
        isWeatherfox = (weatherfox[0] != null);
        var vchecker = Components.classes["@mozilla.org/updates/version-checker;1"].getService(Components.interfaces.nsIVersionChecker);
        
        //not weatherfox
        if (!isWeatherfox) {
          //has to be either 0.5.8 or 0.5.9 
          //so error on lower side or doesn't exist
          if (!file)
            this._doc = this._loadDoc();
          else         
            this._doc = this._transformDoc("0.5.8", false);
                                             
        //is weatherfox
        } else {
          //can't transform a version before 0.4.8
          if (vchecker.compare(weatherfox[0].version, "0.4.8") < 0) {
            this._doc = this._loadDoc();
            this._restart = this._createError(ffIError.SEVERITY_ERROR, "ff.version", null, null);
          } else {
            //transform the weatherfox profile
            this._doc = this._transformDoc(weatherfox[0].version, true);
            var pbs = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService)
            var branch = pbs.getBranch("weatherfox.");
            this._branch.setCharPref("profile.current", branch.getCharPref("profile"));            
          };
          
          em.uninstallExtension("{BA8E053E-2867-4772-B9F0-26A5979AFA09}");
          this._branch.setCharPref("migrated", "*");
          this._restart = this._createError(ffIError.SEVERITY_ERROR, "ff.restart", null, null);
          return;
        };     
      };    
    };       

    //set pref and return
    this._branch.setCharPref("migrated", "0.7.7");
  },
  
  _clearCache: function()
  {
    var dirs = Components.classes["@mozilla.org/file/directory_service;1"].getService(Components.interfaces.nsIProperties);
    var folder = dirs.get("ProfD", Components.interfaces.nsIFile);
            
    //get forecastfox directory
    folder.append("forecastfox");
    if (!folder.exists())
      return;
      
    //make sure its a directory
    if (!folder.isDirectory())
      return;
      
    //get list of files
    var entries = folder.directoryEntries;
    while (entries.hasMoreElements()) {
      var entry = entries.getNext();
      entry = entry.QueryInterface(Components.interfaces.nsIFile);
      var name = entry.leafName;
      
      //remove cache files
      if (name.indexOf("cache") != -1) {
        if (entry.isDirectory())
          entry.remove(true);
        else
          entry.remove(false);
      }; 
    }            
  },
  
  _cacheFiles: function(aFrom, aTo, aDelete)
  {
    var dirs = Components.classes["@mozilla.org/file/directory_service;1"].getService(Components.interfaces.nsIProperties);
    var folder = dirs.get("ProfD", Components.interfaces.nsIFile);
    var i, to, name;
    var fName = "";
    var tName = "";
            
    //get forecastfox directory
    folder.append("forecastfox");
    if (!folder.exists())
      return;
      
    //make sure its a directory
    if (!folder.isDirectory())
      return;
    
    //get names  
    fName = aFrom.getAttribute("id");
    if (aTo)
      tName = aTo.getAttribute("id");
      
    //get a list of cache files
    var entries = folder.directoryEntries;
    var files = new Array();
    while (entries.hasMoreElements()) {
      var entry = entries.getNext();
      entry = entry.QueryInterface(Components.interfaces.nsIFile);
      name = entry.leafName;
      
      //make sure its a cache file
      if (name.indexOf("cache") == -1)
        continue;
        
      //make sure it is a from cache
      if (name.indexOf(fName) != -1)
        files.push(entry);
    };

    //now loop through our files    
    for (i=0; i<files.length; i++) {
      
      //get to file
      if (aTo) {
        to = folder.clone();
        name = files[i].leafName;
        name = name.replace(fName, tName);       
        to.append(name);
        
        //if to alread exists then delete
        if (to.exists())
          to.remove(false);  
                
        //copy from to to
        if (files[i].exists())
          files[i].copyTo(null, to.leafName);
      };
      
      //delete file
      if (aDelete && files[i].exists())
        files[i].remove(false);
    }  
  },
  
  _cachePrefs: function(aFrom, aTo)
  {
    //get a list of preferences
    var prefs = this._default.getChildList("", {});
    var nodes = new Object();
    
    //loop through prefs finding cache prefs
    for (var i=0; i<prefs.length; i++) {
      if (prefs[i].indexOf("cache") == -1)
        continue;
      
      //get profile node that matches
      nodes[prefs[i]] = this._evalPath(aTo, "./pref[@name='" + prefs[i] + "']");
    };

    //replace value and update
    var fName = aFrom.getAttribute("id");
    var tName = aTo.getAttribute("id");
    for (j in nodes) {
      var value = nodes[j].getAttribute("value");
      value = value.replace(fName, tName);
      nodes[j].setAttribute("value", value);
    } 
  },
  
  _transformDoc: function(aFrom, aWeatherfox)
  {
    var processor = Components.classes["@mozilla.org/document-transformer;1?type=xslt"].createInstance(Components.interfaces.nsIXSLTProcessor);  
    var file = null;
    var from = aFrom;
    
    //get the file
    if (!aWeatherfox)
      file = this._getFile("profiles.xml", "exists");
    else
      file = this._getFileWF("profiles.xml", "exists");
      
    //reset the preference branch
    this._removeBranch("forecastfox.", true);
    
    //clear the cache
    this._clearCache();
    
    //copy default if we don't have one
    if (!file) {
      this._restart = this._createError(ffIError.SEVERITY_WARNING, "ff.migrate", null, null);    
      file = this._getDefault();
      return this._getDoc(file);
    };   

    //convert to DOM
    var doc = this._getDoc(file);
          
    //loop up through version
    while (from != "0.7.7") {

      if (!this._validDoc(doc)) {
        this._restart = this._createError(ffIError.SEVERITY_WARNING, "ff.migrate", null, null);    
        file = this._getDefault();
        return this._getDoc(file);
      };
    
      //load the stylesheet
      var sheet = this._getSheet("profiles-" + from + ".xsl");
      if (!sheet) {
        file = this._getDefault();
        return this._getDoc(file);
      };
               
      var sdoc = this._getDoc(sheet);
      if (!this._validDoc(sdoc)) {
        this._restart = this._createError(ffIError.SEVERITY_WARNING, "ff.migrate", null, null);    
        file = this._getDefault();
        return this._getDoc(file);    
      };
    
      //transform document
      processor.reset();
      processor.importStylesheet(sdoc);
      doc = processor.transformToDocument(doc);
           
      //get version doc was transformed to
      var node = this._evalPath(doc, "//profiles/@version");
      if (!node) {
        this._restart = this._createError(ffIError.SEVERITY_WARNING, "ff.migrate", null, null);    
        file = this._getDefault();
        return this._getDoc(file);
      } else
        from = node.textContent;

    
      //flush the doc to disk and then reload it
      this.flush(doc, this._getFile("profiles.xml", "create"));
    };
    
    return doc;
  },
      
  _loadDoc: function()
  {
    //try to load current file
    var isDefault = false;
    var file = this._getFile("profiles.xml", "exists");
    
    //copy default if we don't have one
    if (!file) {
      file = this._getDefault();
      isDefault = true;
    };
    
    //convert to DOM
    var doc = this._getDoc(file);
    if (this._validDoc(doc))
      return doc;
      
    //not successful try again with default
    if (!isDefault) {
      file = this._getDefault();
      doc = this._getDoc(file);
    };
    
    return doc;  
  }, 
   
  _getDoc: function(aFile)
  {
    var parser = Components.classes["@mozilla.org/xmlextras/domparser;1"].createInstance(Components.interfaces.nsIDOMParser);  
    var fiStream = Components.classes["@mozilla.org/network/file-input-stream;1"].createInstance(Components.interfaces.nsIFileInputStream);
    var siStream = Components.classes["@mozilla.org/scriptableinputstream;1"].createInstance(Components.interfaces.nsIScriptableInputStream);
    
    //load document into string variable  
    fiStream.init(aFile, 1, 0, false);
    siStream.init(fiStream);
    var contents = siStream.read(siStream.available());

    //close file
    siStream.close();
    fiStream.close();
    
    //convert string to DOM
    return parser.parseFromString(contents, "text/xml");  
  },
  
  _validDoc: function(aDoc)
  {
    //object passed
    if(!aDoc)
      return false;
      
    //not a parser error
    if (aDoc.childNodes[0].localName == "parsererror")
      return false;
      
    return true;  
  },
      
  _getFile: function(aName, aMode)
  {
    var dirs = Components.classes["@mozilla.org/file/directory_service;1"].getService(Components.interfaces.nsIProperties);
    var folder = dirs.get("ProfD", Components.interfaces.nsIFile);
    
    //get forecastfox directory
    folder.append("forecastfox");
    if (!folder.exists())
      folder.create(folder.DIRECTORY_TYPE, 0755)
    
    //get file
    var file = folder.clone();      
    file.append(aName);
    
    //make sure its not a directory
    if (file.exists()) {   
      if (file.isDirectory()) 
        file.remove(true);
    };

    switch (aMode) {
      case "delete":
        if (file.exists())
          file.remove(false);
        file = null;
        break;
      case "exists":
        if (!file.exists())
          file = null;
        break;
      case "create":
      default:
        if (file.exists())
          file.remove(false);
        file.create(file.NORMAL_FILE_TYPE, 0664);             
        break;
    };  
   
    //make sure file is writable
    if (file) {
      if (!file.isWritable()) 
        file.permissions = 0644;    
    };
    
    //return file
    return file;
  },
      
  _getFileWF: function(aName, aMode)
  {
    var dirs = Components.classes["@mozilla.org/file/directory_service;1"].getService(Components.interfaces.nsIProperties);
    var folder = dirs.get("ProfD", Components.interfaces.nsIFile);
    
    //get forecastfox directory
    folder.append("weatherfox");
    if (!folder.exists())
      folder.create(folder.DIRECTORY_TYPE, 0755)
    
    //get file
    var file = folder.clone();      
    file.append(aName);
    
    //make sure its not a directory
    if (file.exists()) {   
      if (file.isDirectory()) 
        file.remove(true);
    };

    switch (aMode) {
      case "delete":
        if (file.exists())
          file.remove(false);
        file = null;
        break;
      case "exists":
        if (!file.exists())
          file = null;
        break;
      case "create":
      default:
        if (file.exists())
          file.remove(false);
        file.create(file.NORMAL_FILE_TYPE, 0664);             
        break;
    };  
   
    //make sure file is writable
    if (file) {
      if (!file.isWritable()) 
        file.permissions = 0644;    
    };
    
    //return file
    return file;
  },
      
  _getDefault: function()
  {
    var dirs = Components.classes["@mozilla.org/file/directory_service;1"].getService(Components.interfaces.nsIProperties);
    var profile = dirs.get("ProfD", Components.interfaces.nsIFile);
    var app = dirs.get("XCurProcD", Components.interfaces.nsIFile);
    
    //reset the branch and directory
    this._removeDir("forecastfox", true);
    this._removeBranch("forecastfox.", true);
            
    //first try profile path
    var file = profile.clone();
    file.append("extensions");
    file.append("{0538E3E3-7E9B-4d49-8831-A227C80A7AD3}");
    file.append("defaults");
    file.append("profiles.xml");
      
    //then try global
    if (!file.exists()) {
      file = app.clone(); 
      file.append("extensions");
      file.append("{0538E3E3-7E9B-4d49-8831-A227C80A7AD3}");
      file.append("defaults");
      file.append("profiles.xml");
    };
      
    //if neither then error
    if (!file.exists())
      return null;
         
    var folder = profile.clone();
    folder.append("forecastfox");
    
    file.copyTo(folder, "profiles.xml");
      
    //set profile and return      
    this._current = "default"; 
    return file;              
  },
  
  _getSheet: function(aName)
  {
    var dirs = Components.classes["@mozilla.org/file/directory_service;1"].getService(Components.interfaces.nsIProperties);
    var profile = dirs.get("ProfD", Components.interfaces.nsIFile);
    var app = dirs.get("XCurProcD", Components.interfaces.nsIFile);
      
    //first try profile path
    var file = profile.clone();
    file.append("extensions");
    file.append("{0538E3E3-7E9B-4d49-8831-A227C80A7AD3}");
    file.append("defaults");
    file.append("transforms")
    file.append(aName);
      
    //then try global
    if (!file.exists()) {
      file = app.clone(); 
      file.append("extensions");
      file.append("{0538E3E3-7E9B-4d49-8831-A227C80A7AD3}");
      file.append("defaults");
      file.append("transforms")      
      file.append(aName);
    };
      
    //if neither then error
    if (!file.exists())
      return null;

    return file;              
  },
  
  _getExport: function()
  {
    var dirs = Components.classes["@mozilla.org/file/directory_service;1"].getService(Components.interfaces.nsIProperties);
    var profile = dirs.get("ProfD", Components.interfaces.nsIFile);
    var app = dirs.get("XCurProcD", Components.interfaces.nsIFile);
      
    //first try profile path
    var file = profile.clone();
    file.append("extensions");
    file.append("{0538E3E3-7E9B-4d49-8831-A227C80A7AD3}");
    file.append("defaults");
    file.append("export.xml");
      
    //then try global
    if (!file.exists()) {
      file = app.clone(); 
      file.append("extensions");
      file.append("{0538E3E3-7E9B-4d49-8831-A227C80A7AD3}");
      file.append("defaults");
      file.append("export.xml");
    };
      
    //if neither then error
    if (!file.exists())
      return null;

    return file;              
  },
      
  _removeDir: function(aDir, aRecreate)
  {
    //get directory
    var dirs = Components.classes["@mozilla.org/file/directory_service;1"].getService(Components.interfaces.nsIProperties);  
    var folder = dirs.get("ProfD", Components.interfaces.nsIFile);
    folder.append(aDir);
    
    //delete it
    if (folder.exists())
      folder.remove(true);
      
    //recreate directory
    if (aRecreate)
      folder.create(folder.DIRECTORY_TYPE, 0755);    
  },
  
  _removeBranch: function(aBranch, aReset)
  {
    var pbs = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService)
    var branch = pbs.getBranch(aBranch);

    //do a reset
    if (aReset) {   
      //get list of prefs
      var prefs = branch.getChildList("", {});
      
      for (var x=0; x<prefs.length; x++) {    
        //clear user value
        if (branch.prefHasUserValue(prefs[x])) {        
          if (this._excludes[prefs[x]])
            continue;
          else
            branch.clearUserPref(prefs[x]);  
        };
      };
    } else    
      branch.deleteBranch("");
  },

  _stripName: function(name)
  {
    // remove illegal characters
    return name.replace(/[\\\/:*?<>|\'\"]/g, "")
  },
  
  ////////////////////////////
  // nsIObserver  
  observe: function(aSubject, aTopic, aData) 
  {
    var method = null;
    var data = null;
    if (aData.indexOf(",") != -1) {
      method = aData.substring(0, aData.indexOf(","));
      data = aData.substring(aData.indexOf(",") + 1, aData.length);
    } else
      method = aData.substring(0, aData.length);
    switch (method) {
      case "start":
        if (this._groups[data] == null)
          this._groups[data] = new Object();
        this._groups[data].loading = true;
        break;
      case "error":
      case "end":
        if (this._groups[data] == null)
          this._groups[data] = new Object();
        this._groups[data].loading = false;
        this._releaseLocks();
        break;
    }
  },
  
  ///////////////////////////
  // nsITimerCallback
  notify: function(timer)
  {
    var profiles = this._evalPath(this._doc, "//profiles");  
    profiles = profiles.getElementsByTagName("profile");
    
    // go to next profile in the list
    for (var x=0; x<profiles.length; x++) {
      if (profiles[x].getAttribute("id") == this._current) {
        this._branch.setCharPref("profile.current", profiles[( x+1 < profiles.length ) ? x+1 : 0 ].getAttribute("id"));
        break;
      };
    };
    
    //schedule rotation
    this._schedule();
  },
       
  ///////////////////////////
  // nsISupports
  QueryInterface: function (aIID)
  {
    if (!aIID.equals(Components.interfaces.ffIProfiles) &&
        !aIID.equals(Components.interfaces.nsISupports))
      throw Components.results.NS_ERROR_NO_INTERFACE;
    return this;
  }
};

/******************************************************************************
 * XPCOM Functions for construction and registration
 ******************************************************************************/
var gModule = {
  _firstTime: true,
  registerSelf: function(aCompMgr, aFileSpec, aLocation, aType)
  {
    if (this._firstTime) {
      this._firstTime = false;
      throw Components.results.NS_ERROR_FACTORY_REGISTER_AGAIN;
    };
    aCompMgr = aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.registerFactoryLocation(CLASS_ID, CLASS_NAME, CONTRACT_ID, aFileSpec, aLocation, aType);
  },

  unregisterSelf: function(aCompMgr, aLocation, aType)
  {
    aCompMgr = aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.unregisterFactoryLocation(CLASS_ID, aLocation);        
  },
  
  getClassObject: function(aCompMgr, aCID, aIID)
  {
    if (!aIID.equals(Components.interfaces.nsIFactory))
      throw Components.results.NS_ERROR_NOT_IMPLEMENTED;

    if (aCID.equals(CLASS_ID))
      return gFactory;

    throw Components.results.NS_ERROR_NO_INTERFACE;
  },

  canUnload: function(aCompMgr) { return true; }
};

var gFactory = {
  createInstance: function (aOuter, aIID)
  {
    if (aOuter != null)
      throw Components.results.NS_ERROR_NO_AGGREGATION;
    return (new ffProfiles()).QueryInterface(aIID);
  }
};

function NSGetModule(aCompMgr, aFileSpec) { return gModule; }